import numpy as np
import torch
import torchvision
import torchvision.transforms as transforms
import torch.optim as optim
from PIL import Image
from torchvision.utils import make_grid

import matplotlib.pyplot as plt
import numpy as np

plt.rcParams.update({'font.size': 22})

# feature-map-visualization
# https://www.kaggle.com/code/subhajeetdas/pytorch-feature-map-filters-visualization
# https://discuss.pytorch.org/t/visualize-feature-map/29597/45
# https://ravivaishnav20.medium.com/visualizing-feature-maps-using-pytorch-12a48cd1e573

IMG_SIZE = 64
BATH_SIZE = 8

class_names = {
    0: "Speed limit (20km/h)",
    1: "Speed limit (30km/h)",
    2: "Speed limit (50km/h)",
    3: "Speed limit (60km/h)",
    4: "Speed limit (70km/h)",
    5: "Speed limit (80km/h)",
    6: "End of speed limit (80km/h)",
    7: "Speed limit (100km/h)",
    8: "Speed limit (120km/h)",
    9: "No passing",
    10: "No passing veh over 3.5 tons",
    11: "Right-of-way at intersection",
    12: "Priority road",
    13: "Yield",
    14: "Stop",
    15: "No vehicles",
    16: "Veh > 3.5 tons prohibited",
    17: "No entry",
    18: "General caution",
    19: "Dangerous curve left",
    20: "Dangerous curve right",
    21: "Double curve",
    22: "Bumpy road",
    23: "Slippery road",
    24: "Road narrows on the right",
    25: "Road work",
    26: "Traffic signals",
    27: "Pedestrians",
    28: "Children crossing",
    29: "Bicycles crossing",
    30: "Beware of ice/snow",
    31: "Wild animals crossing",
    32: "End speed + passing limits",
    33: "Turn right ahead",
    34: "Turn left ahead",
    35: "Ahead only",
    36: "Go straight or right",
    37: "Go straight or left",
    38: "Keep right",
    39: "Keep left",
    40: "Roundabout mandatory",
    41: "End of no passing",
    42: "End no passing veh > 3.5 tons"
}

def my_visu(feature_map):
    
    figure = plt.figure(figsize=(8, 8))
    print("my_visu feature_map.shape", feature_map.shape)
    feature_map = feature_map.squeeze(0)
    print("feature_map", feature_map.shape)
    for idx in range(16):
        figure.add_subplot(4, 4, idx+1)
        img = feature_map[idx, :, :].detach().numpy()
        plt.imshow(img, cmap="gray")
    plt.show()    

transform = transforms.Compose(
    [
     transforms.Grayscale(),
     transforms.ToTensor(),
     transforms.Resize((IMG_SIZE, IMG_SIZE), antialias=True),
     ])

trainset = torchvision.datasets.GTSRB(root='./data', split="train",
                                        download=True, transform=transform)

testset = torchvision.datasets.GTSRB(root='./data', split="test",
                                       download=True, transform=transform)


figure = plt.figure(figsize=(8, 8))
for i in range(1, 26):
    img, label = trainset[np.random.randint(0, len(trainset))]
    figure.add_subplot(5, 5, i)
    print(img.size())
    img_t = np.transpose(img, (1, 2, 0))
    print(type(img))
    #img_t = torch.squeeze(img, 1)
    print(img_t.size())
    plt.title(class_names.get(label, "Unknown class ID"))
    plt.axis("off")
    plt.imshow(img_t, cmap="gray")    
    print(i)
plt.show()


testloader = torch.utils.data.DataLoader(testset, batch_size=BATH_SIZE,
                                         shuffle=False, num_workers=1)

trainloader = torch.utils.data.DataLoader(trainset, batch_size=BATH_SIZE,
                                          shuffle=True, num_workers=1)

import torch.nn as nn
import torch.nn.functional as F


class Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 16, 5, 1, 2)
        self.conv2 = nn.Conv2d(16, 32, 5, 1, 2)
        self.conv3 = nn.Conv2d(32, 48, 3, 1, 1)
        #self.conv4 = nn.Conv2d(64, 128, 3, 1, 1)
        self.pool = nn.MaxPool2d(2, 2)
        self.fc1 = nn.LazyLinear(120)
        self.fc2 = nn.LazyLinear(80)
        self.fc3 = nn.LazyLinear(43)

    def forward(self, x):
        self.feature_map1 = self.pool(F.relu(self.conv1(x)))
        self.feature_map2 = self.pool(F.relu(self.conv2(self.feature_map1)))
        self.feature_map3 = self.pool(F.relu(self.conv3(self.feature_map2)))
        x = torch.flatten(self.feature_map3, 1) # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

net = Net() #torchvision.models.alexnet(pretrained=True)     

criterion = torch.nn.CrossEntropyLoss()
optimizer = optim.SGD(net.parameters(), lr=0.01)

for epoch in range(2):  # loop over the dataset multiple times
    
    running_loss = 0.0
    for i, data in enumerate(trainloader, 0):
        # get the inputs; data is a list of [inputs, labels]
        inputs, labels = data

        # zero the parameter gradients
        optimizer.zero_grad()

        # forward + backward + optimize
        outputs = net(inputs)
        loss = criterion(outputs, labels)
        loss.backward()
        optimizer.step()

        # print statistics
        running_loss += loss.item()
        if i % 20 == 19:    # print every 20 mini-batches
            print(f'[{epoch + 1}, {i + 1:5d}] loss: {running_loss / 20:.3f}')
            running_loss = 0.0
            
print('Finished Training')

image, _ = testset[0]
image.unsqueeze_(0)
net(image)
print(net.feature_map1.shape)
my_visu(net.feature_map1)
my_visu(net.feature_map2)
my_visu(net.feature_map3)


